@@ -1,4 +1,4 @@
-use std::{ops::Deref, usize};
+use std::{collections::HashMap, ops::Deref, usize};
 
 use crate::{SimpleDnsError, OPCODE};
 
@@ -9,21 +9,28 @@ use super::{DnsPacketContent, PacketHeader, Question, ResourceRecord};
 /// You have to build the packet in order.  
 /// ex: It is not possible to add a question after an answer
 pub struct PacketBuf {
+    name_refs: HashMap<u64, usize>,
     inner: Vec<u8>,
+    compression: bool,
 }
 
 impl PacketBuf {
     /// Creates a new empty PacketBuf
-    pub fn new(header: PacketHeader) -> Self {
+    pub fn new(header: PacketHeader, compression: bool) -> Self {
         let mut inner = vec![0; 12];
         header.write_to(&mut inner);
-        Self { inner }
+        let name_refs = HashMap::new();
+        Self {
+            inner,
+            name_refs,
+            compression,
+        }
     }
 
     /// Creates a new empty PacketBuf with a query header
-    pub fn new_query() -> Self {
+    pub fn new_query(compression: bool) -> Self {
         let header = PacketHeader::new_query(0, false);
-        Self::new(header)
+        Self::new(header, compression)
     }
 
     /// Add a [`Question`] to this packet.  
@@ -32,8 +39,11 @@ impl PacketBuf {
         if self.has_answers() || self.has_name_servers() || self.has_additional_records() {
             return Err(SimpleDnsError::InvalidDnsPacket);
         }
-
-        question.append_to_vec(&mut self.inner)?;
+        if self.compression {
+            question.compress_append_to_vec(&mut self.inner, &mut self.name_refs)?;
+        } else {
+            question.append_to_vec(&mut self.inner)?;
+        }
         let questions_count = PacketHeader::read_questions(&self.inner);
         PacketHeader::write_questions(&mut self.inner, questions_count + 1);
         Ok(())
@@ -46,7 +56,12 @@ impl PacketBuf {
             return Err(SimpleDnsError::InvalidDnsPacket);
         }
 
-        answer.append_to_vec(&mut self.inner)?;
+        if self.compression {
+            answer.compress_append_to_vec(&mut self.inner, &mut self.name_refs)?;
+        } else {
+            answer.append_to_vec(&mut self.inner)?;
+        }
+
         let answers_count = PacketHeader::read_answers(&self.inner);
         PacketHeader::write_answers(&mut self.inner, answers_count + 1);
         Ok(())
@@ -59,7 +74,12 @@ impl PacketBuf {
             return Err(SimpleDnsError::InvalidDnsPacket);
         }
 
-        name_server.append_to_vec(&mut self.inner)?;
+        if self.compression {
+            name_server.compress_append_to_vec(&mut self.inner, &mut self.name_refs)?;
+        } else {
+            name_server.append_to_vec(&mut self.inner)?;
+        }
+
         let ns_count = PacketHeader::read_name_servers(&self.inner);
         PacketHeader::write_name_servers(&mut self.inner, ns_count + 1);
         Ok(())
@@ -70,7 +90,11 @@ impl PacketBuf {
         &mut self,
         additional_record: &ResourceRecord,
     ) -> crate::Result<()> {
-        additional_record.append_to_vec(&mut self.inner)?;
+        if self.compression {
+            additional_record.compress_append_to_vec(&mut self.inner, &mut self.name_refs)?;
+        } else {
+            additional_record.append_to_vec(&mut self.inner)?;
+        }
         let additional_records_count = PacketHeader::read_additional_records(&self.inner);
         PacketHeader::write_additional_records(&mut self.inner, additional_records_count + 1);
         Ok(())
@@ -107,12 +131,11 @@ impl PacketBuf {
     }
 
     /// Returns an Iterator over questions of this packet
-    pub fn questions_iter(&self) -> PacketSectionIter<Question> {
+    pub fn questions_iter(&self) -> QuestionsIter {
         let total = PacketHeader::read_questions(&self.inner);
         let pos = 12;
 
-        PacketSectionIter {
-            _marker: std::marker::PhantomData::default(),
+        QuestionsIter {
             buf: self,
             total,
             pos,
@@ -126,7 +149,9 @@ impl PacketBuf {
 impl From<&[u8]> for PacketBuf {
     fn from(buffer: &[u8]) -> Self {
         Self {
+            name_refs: HashMap::new(),
             inner: buffer.to_vec(),
+            compression: true,
         }
     }
 }
@@ -141,28 +166,21 @@ impl Deref for PacketBuf {
 
 /// Iterate over the questions of a [`PacketBuf`]
 /// If a question is not valid, the iterator will stop
-pub struct PacketSectionIter<'a, T>
-where
-    T: DnsPacketContent<'a>,
-{
-    _marker: std::marker::PhantomData<&'a T>,
+pub struct QuestionsIter<'a> {
     buf: &'a PacketBuf,
     curr: u16,
     total: u16,
     pos: usize,
 }
-impl<'a, T> Iterator for PacketSectionIter<'a, T>
-where
-    T: DnsPacketContent<'a>,
-{
-    type Item = T;
+impl<'a> Iterator for QuestionsIter<'a> {
+    type Item = Question<'a>;
 
     fn next(&mut self) -> Option<Self::Item> {
         if self.curr >= self.total {
             return None;
         }
 
-        let question = Self::Item::parse(&self.buf, self.pos).ok()?;
+        let question = Self::Item::parse(self.buf, self.pos).ok()?;
         self.curr += 1;
         self.pos += question.len();
 
@@ -258,25 +276,41 @@ impl<'a> Packet<'a> {
         Self::add_section(&mut out, &self.answers)?;
         Self::add_section(&mut out, &self.name_servers)?;
         Self::add_section(&mut out, &self.additional_records)?;
+        self.write_header(&mut out);
 
+        Ok(out)
+    }
+
+    /// Creates a new [Vec`<u8>`](`Vec<T>`) from the contents of this package with [Name](`Name`) compression
+    pub fn build_bytes_vec_compressed(&self) -> crate::Result<Vec<u8>> {
+        let mut out = vec![0u8; 12];
+        let mut name_refs = HashMap::new();
+
+        Self::add_section_compressed(&mut out, &mut name_refs, &self.questions)?;
+        Self::add_section_compressed(&mut out, &mut name_refs, &self.answers)?;
+        Self::add_section_compressed(&mut out, &mut name_refs, &self.name_servers)?;
+        Self::add_section_compressed(&mut out, &mut name_refs, &self.additional_records)?;
+        self.write_header(&mut out);
+
+        Ok(out)
+    }
+    fn write_header(&self, out: &mut Vec<u8>) {
         self.header.write_to(&mut out[0..12]);
         if !self.questions.is_empty() {
-            PacketHeader::write_questions(&mut out, self.questions.len() as u16)
+            PacketHeader::write_questions(out, self.questions.len() as u16)
         }
 
         if !self.answers.is_empty() {
-            PacketHeader::write_answers(&mut out, self.answers.len() as u16)
+            PacketHeader::write_answers(out, self.answers.len() as u16)
         }
 
         if !self.name_servers.is_empty() {
-            PacketHeader::write_name_servers(&mut out, self.name_servers.len() as u16)
+            PacketHeader::write_name_servers(out, self.name_servers.len() as u16)
         }
 
         if !self.additional_records.is_empty() {
-            PacketHeader::write_additional_records(&mut out, self.additional_records.len() as u16)
+            PacketHeader::write_additional_records(out, self.additional_records.len() as u16)
         }
-
-        Ok(out)
     }
 
     fn add_section<'b, T: DnsPacketContent<'b>>(
@@ -289,6 +323,16 @@ impl<'a> Packet<'a> {
 
         Ok(())
     }
+    fn add_section_compressed<'b, T: DnsPacketContent<'b>>(
+        out: &mut Vec<u8>,
+        name_refs: &mut HashMap<u64, usize>,
+        section: &[T],
+    ) -> crate::Result<()> {
+        for item in section {
+            item.compress_append_to_vec(out, name_refs)?;
+        }
+        Ok(())
+    }
 }
 
 #[cfg(test)]
@@ -322,6 +366,8 @@ mod tests {
 
         let parsed = parsed.unwrap();
         assert_eq!(2, parsed.questions.len());
+        assert_eq!("_srv._udp.local", parsed.questions[0].qname.to_string());
+        assert_eq!("_srv2._udp.local", parsed.questions[1].qname.to_string());
     }
 
     #[test]
@@ -369,7 +415,7 @@ mod tests {
 
     #[test]
     fn bufpacket_add_question() {
-        let mut buf_packet = PacketBuf::new(PacketHeader::new_query(0, false));
+        let mut buf_packet = PacketBuf::new(PacketHeader::new_query(0, false), false);
         let question = Question::new(
             Name::new_unchecked("_srv._udp.local"),
             QTYPE::TXT,
@@ -386,17 +432,17 @@ mod tests {
         assert!(buf_packet.add_question(&question).is_ok());
         assert!(buf_packet.has_questions());
 
-        let mut buf_packet = PacketBuf::new(PacketHeader::new_query(0, false));
+        let mut buf_packet = PacketBuf::new(PacketHeader::new_query(0, false), false);
         buf_packet.add_answer(&resource).unwrap();
         assert!(buf_packet.add_question(&question).is_err());
         assert!(!buf_packet.has_questions());
 
-        let mut buf_packet = PacketBuf::new(PacketHeader::new_query(0, false));
+        let mut buf_packet = PacketBuf::new(PacketHeader::new_query(0, false), false);
         buf_packet.add_name_server(&resource).unwrap();
         assert!(buf_packet.add_question(&question).is_err());
         assert!(!buf_packet.has_questions());
 
-        let mut buf_packet = PacketBuf::new(PacketHeader::new_query(0, false));
+        let mut buf_packet = PacketBuf::new(PacketHeader::new_query(0, false), false);
         buf_packet.add_additional_record(&resource).unwrap();
         assert!(buf_packet.add_question(&question).is_err());
         assert!(!buf_packet.has_questions());
@@ -404,7 +450,7 @@ mod tests {
 
     #[test]
     fn bufpacket_add_answers() {
-        let mut buf_packet = PacketBuf::new(PacketHeader::new_query(0, false));
+        let mut buf_packet = PacketBuf::new(PacketHeader::new_query(0, false), false);
         let resource = ResourceRecord::new(
             Name::new_unchecked("srv._udp.local"),
             CLASS::IN,
@@ -415,12 +461,12 @@ mod tests {
         assert!(buf_packet.add_answer(&resource).is_ok());
         assert!(buf_packet.has_answers());
 
-        let mut buf_packet = PacketBuf::new(PacketHeader::new_query(0, false));
+        let mut buf_packet = PacketBuf::new(PacketHeader::new_query(0, false), false);
         buf_packet.add_name_server(&resource).unwrap();
         assert!(buf_packet.add_answer(&resource).is_err());
         assert!(!buf_packet.has_answers());
 
-        let mut buf_packet = PacketBuf::new(PacketHeader::new_query(0, false));
+        let mut buf_packet = PacketBuf::new(PacketHeader::new_query(0, false), false);
         buf_packet.add_additional_record(&resource).unwrap();
         assert!(buf_packet.add_answer(&resource).is_err());
         assert!(!buf_packet.has_answers());
@@ -428,7 +474,7 @@ mod tests {
 
     #[test]
     fn bufpacket_add_name_servers() {
-        let mut buf_packet = PacketBuf::new(PacketHeader::new_query(0, false));
+        let mut buf_packet = PacketBuf::new(PacketHeader::new_query(0, false), false);
         let resource = ResourceRecord::new(
             Name::new_unchecked("_srv._udp.local"),
             CLASS::IN,
@@ -439,7 +485,7 @@ mod tests {
         assert!(buf_packet.add_name_server(&resource).is_ok());
         assert!(buf_packet.has_name_servers());
 
-        let mut buf_packet = PacketBuf::new(PacketHeader::new_query(0, false));
+        let mut buf_packet = PacketBuf::new(PacketHeader::new_query(0, false), false);
         buf_packet.add_additional_record(&resource).unwrap();
         assert!(buf_packet.add_name_server(&resource).is_err());
         assert!(!buf_packet.has_name_servers());
@@ -447,7 +493,7 @@ mod tests {
 
     #[test]
     fn bufpacket_add_additional_records() {
-        let mut buf_packet = PacketBuf::new(PacketHeader::new_query(0, false));
+        let mut buf_packet = PacketBuf::new(PacketHeader::new_query(0, false), false);
         let resource = ResourceRecord::new(
             Name::new_unchecked("_srv._udp.local"),
             CLASS::IN,
@@ -461,7 +507,7 @@ mod tests {
 
     #[test]
     fn bufpacket_questions_iter() {
-        let mut buf_packet = PacketBuf::new(PacketHeader::new_query(0, false));
+        let mut buf_packet = PacketBuf::new(PacketHeader::new_query(0, false), false);
         let question = Question::new(
             Name::new_unchecked("_srv._udp.local"),
             QTYPE::TXT,
@@ -472,7 +518,6 @@ mod tests {
         buf_packet.add_question(&question).unwrap();
         buf_packet.add_question(&question).unwrap();
 
-        let questions: Vec<Question> = buf_packet.questions_iter().collect();
-        assert_eq!(2, questions.len());
+        assert_eq!(2, buf_packet.questions_iter().count());
     }
 }
