@@ -1,5 +1,6 @@
 use super::{CharacterString, DnsPacketContent, Name, TYPE};
 use core::fmt::Debug;
+use std::collections::HashMap;
 
 mod a;
 mod aaaa;
@@ -13,6 +14,7 @@ mod wks;
 
 pub use a::A;
 pub use aaaa::AAAA;
+use byteorder::{BigEndian, ByteOrder};
 pub use hinfo::HINFO;
 pub use minfo::MINFO;
 pub use mx::MX;
@@ -21,6 +23,7 @@ pub use soa::SOA;
 pub use srv::SRV;
 pub use wks::WKS;
 
+/// Represents the RData of each [`TYPE`]
 #[derive(Debug, Eq, PartialEq, Hash)]
 pub enum RData<'a> {
     A(A),
@@ -43,31 +46,18 @@ pub enum RData<'a> {
     NULL(u16, NULL<'a>),
 }
 
-impl<'a> RData<'a> {
-    pub(crate) fn len(&self) -> usize {
-        match &self {
-            RData::A(data) => data.len(),
-            RData::AAAA(data) => data.len(),
-            RData::NS(data)
-            | RData::CNAME(data)
-            | RData::MB(data)
-            | RData::MG(data)
-            | RData::MR(data)
-            | RData::PTR(data)
-            | RData::MF(data)
-            | RData::MD(data) => data.len(),
-            RData::HINFO(data) => data.len(),
-            RData::MINFO(data) => data.len(),
-            RData::MX(data) => data.len(),
-            RData::NULL(_, data) => data.len(),
-            RData::TXT(data) => data.len(),
-            RData::SOA(data) => data.len(),
-            RData::WKS(data) => data.len(),
-            RData::SRV(data) => data.len(),
-        }
+impl<'a> DnsPacketContent<'a> for RData<'a> {
+    fn parse(data: &'a [u8], position: usize) -> crate::Result<Self>
+    where
+        Self: Sized,
+    {
+        let rdatatype = BigEndian::read_u16(&data[position..position + 2]).into();
+        let rdatalen = BigEndian::read_u16(&data[position + 8..position + 10]) as usize;
+
+        parse_rdata(&data[position + 10..position + 10 + rdatalen], 0, rdatatype)
     }
 
-    pub(crate) fn append_to_vec(&self, out: &mut Vec<u8>) -> crate::Result<()> {
+    fn append_to_vec(&self, out: &mut Vec<u8>) -> crate::Result<()> {
         match &self {
             RData::A(data) => data.append_to_vec(out),
             RData::AAAA(data) => data.append_to_vec(out),
@@ -90,6 +80,59 @@ impl<'a> RData<'a> {
         }
     }
 
+    fn compress_append_to_vec(
+        &self,
+        out: &mut Vec<u8>,
+        name_refs: &mut HashMap<u64, usize>,
+    ) -> crate::Result<()> {
+        match &self {
+            RData::A(data) => data.compress_append_to_vec(out, name_refs),
+            RData::AAAA(data) => data.compress_append_to_vec(out, name_refs),
+            RData::NS(data)
+            | RData::CNAME(data)
+            | RData::MB(data)
+            | RData::MG(data)
+            | RData::MR(data)
+            | RData::PTR(data)
+            | RData::MF(data)
+            | RData::MD(data) => data.compress_append_to_vec(out, name_refs),
+            RData::HINFO(data) => data.compress_append_to_vec(out, name_refs),
+            RData::MINFO(data) => data.compress_append_to_vec(out, name_refs),
+            RData::MX(data) => data.compress_append_to_vec(out, name_refs),
+            RData::NULL(_, data) => data.compress_append_to_vec(out, name_refs),
+            RData::TXT(data) => data.compress_append_to_vec(out, name_refs),
+            RData::SOA(data) => data.compress_append_to_vec(out, name_refs),
+            RData::WKS(data) => data.compress_append_to_vec(out, name_refs),
+            RData::SRV(data) => data.compress_append_to_vec(out, name_refs),
+        }
+    }
+
+    fn len(&self) -> usize {
+        match &self {
+            RData::A(data) => data.len(),
+            RData::AAAA(data) => data.len(),
+            RData::NS(data)
+            | RData::CNAME(data)
+            | RData::MB(data)
+            | RData::MG(data)
+            | RData::MR(data)
+            | RData::PTR(data)
+            | RData::MF(data)
+            | RData::MD(data) => data.len(),
+            RData::HINFO(data) => data.len(),
+            RData::MINFO(data) => data.len(),
+            RData::MX(data) => data.len(),
+            RData::NULL(_, data) => data.len(),
+            RData::TXT(data) => data.len(),
+            RData::SOA(data) => data.len(),
+            RData::WKS(data) => data.len(),
+            RData::SRV(data) => data.len(),
+        }
+    }
+}
+
+impl<'a> RData<'a> {
+    /// Returns the [`TYPE`] of this RData
     pub fn type_code(&self) -> TYPE {
         match self {
             RData::A(_) => TYPE::A,
@@ -114,7 +157,7 @@ impl<'a> RData<'a> {
     }
 }
 
-pub(crate) fn parse_rdata(data: &[u8], position: usize, rdatatype: TYPE) -> crate::Result<RData> {
+fn parse_rdata(data: &[u8], position: usize, rdatatype: TYPE) -> crate::Result<RData> {
     let rdata = match rdatatype {
         TYPE::A => RData::A(A::parse(data, position)?),
         TYPE::AAAA => RData::AAAA(AAAA::parse(data, position)?),
